import socketserver
import struct

from .socks5 import Socks5Client

# from helpers import hexstring


class DNSQueryHandler(socketserver.BaseRequestHandler):
    """UDP DNS Proxy handler"""

    def handle(self):
        data = self.request[0].strip()
        sock = self.request[1]

        sc = Socks5Client(
            self.local_addr, data=(self.remote_tcp_dns, 53), enable_ssl=False
        )
        server = sc.connect()
        server.sendall(struct.pack("!H", len(data)) + data)
        result = server.recv(65535)
        server.close()
        sock.sendto(result[2:], self.client_address)



import logging
import select
import socket
import socketserver
import ssl
import struct

from .helpers import hexstring, make_connection, my_inet_aton, trigger_upstream_check

ssl_context = ssl.create_default_context()
ssl_context.check_hostname = False
ssl_context.verify_mode = ssl.CERT_NONE

####################################
# Constants
####################################

BUF_SIZE = 1024
TIME_OUT = 30

# Socks5 stages
INIT_STAGE = 0
AUTH_STAGE = 1
FINAL_STAGE = 2
CONN_ACCEPTED = 3

# Socks5 auth codes
AUTH_SUCCESSFUL = b"\x00"
AUTH_ERR_SERVER = b"\x01"
AUTH_ERR_BANDWIDTH = b"\x02"
AUTH_ERR_NOPLANFOUND = b"\x03"
AUTH_ERR_USERNOTFOUND = b"\x04"


####################################
# Exceptions
####################################


class Socks5Exception(Exception):
    """Base socks5 exception class"""

    pass


class Socks5NoAuthMethodAccepted(Socks5Exception):
    def __init__(self):
        Exception.__init__(self, "No auth method accepted.")


class Socks5AuthFailed(Socks5Exception):
    def __init__(self, reason=None):
        if reason:
            Exception.__init__(self, "Authentication failed: %s." % reason)
        else:
            Exception.__init__(self, "Authentication failed.")


class Socks5DnsFailed(Socks5Exception):
    def __init__(self):
        Exception.__init__(self, "DNS resolve failed.")


class Socks5ConnectionFailed(Socks5Exception):
    def __init__(self):
        Exception.__init__(self, "Connection to upstream/destination failed.")


class Socks5RemoteConnectionClosed(Socks5Exception):
    def __init__(self):
        Exception.__init__(self, "Remote connection closed.")


class Socks5SocketError(Socks5Exception):
    def __init__(self):
        Exception.__init__(self, "A socket error occurred when forwarding.")


class Socks5ConnectionClosed(Socks5Exception):
    def __init__(self):
        Exception.__init__(self, "Socks5 connection closed.")


class Socks5NotImplemented(Socks5Exception):
    def __init__(self):
        Exception.__init__(self, "Protocol not implemented yet.")


class Socks5PortForbidden(Socks5Exception):
    def __init__(self, port):
        Exception.__init__(self, "Port %d is not allowed" % port)


####################################
# Socks5 handlers
####################################


class Socks5RequestHandler(socketserver.StreamRequestHandler):
    """Socks5 request handler"""

    def setup(self):
        socketserver.StreamRequestHandler.setup(self)

        self.bytes_in = 0
        self.bytes_out = 0

        self.member_id = 0

        self.client_name = None
        self.server_name = None

    def handle(self):
        """Main handler"""

        stage = INIT_STAGE
        leftover = b""
        dest = None

        try:
            while stage < CONN_ACCEPTED:
                data = self.request.recv(BUF_SIZE)

                # Client closed connection
                if not data:
                    raise Socks5ConnectionClosed

                data = leftover + data

                if len(data) < 3:
                    leftover = data
                    continue

                # Init stage
                if stage == INIT_STAGE:
                    # If no auth required
                    if not self.local_auth and data == b"\x05\x01\x00":
                        self.request.sendall(b"\x05\x00")
                        stage = FINAL_STAGE
                        continue
                    # if username/password auth required
                    elif self.local_auth and data == b"\x05\x01\x02":
                        self.request.sendall(b"\x05\x02")
                        stage = AUTH_STAGE
                        continue
                    # no auth method accepted
                    else:
                        self.request.sendall(b"\x05\xff")
                        # print(hexstring(data))
                        raise Socks5NoAuthMethodAccepted

                # Auth stage
                elif stage == AUTH_STAGE:
                    (name_length,) = struct.unpack("B", data[1:2])
                    if len(data[2:]) < name_length + 1:
                        leftover = data
                        continue
                    (pass_length,) = struct.unpack(
                        "B", data[2 + name_length : 2 + name_length + 1]
                    )
                    if len(data[2 + name_length + 1 :]) < pass_length:
                        leftover = data
                        continue

                    username = data[2 : 2 + name_length].decode("utf-8")
                    password = data[2 + name_length + 1 :].decode("utf-8")

                    self.member_id, error_code = self.authority.auth(username, password)

                    if error_code != AUTH_SUCCESSFUL:
                        self.request.sendall(b"\x01" + error_code)
                        logging.info("Auth failed for user: %s", username)
                        raise Socks5AuthFailed
                    else:
                        self.request.sendall(b"\x01\x00")
                        logging.info("Auth succeeded for user: %s", username)
                        stage = FINAL_STAGE

                # Final stage
                elif stage == FINAL_STAGE:
                    if len(data) < 10:
                        leftover = data
                        continue
                    # Only TCP connections and IPV4 are allowed
                    if data[:2] != b"\x05\x01" or data[3:4] == b"\x04":
                        # Protocol error
                        self.request.sendall(b"\x05\x07")
                        raise Socks5NotImplemented
                    else:
                        domain = port = None
                        # Connect by domain name
                        if data[3:4] == b"\x03" or data[3:4] == b"\x02":
                            (length,) = struct.unpack("B", data[4:5])
                            domain = data[5 : 5 + length]
                            (port,) = struct.unpack("!H", data[5 + length :])
                        # Connect by ip address
                        elif data[3:4] == b"\x01":
                            domain = socket.inet_ntoa(data[4:8])
                            (port,) = struct.unpack("!H", data[8:])
                        try:
                            # Resolve domain to ip
                            if data[3:4] == b"\x02":
                                _, _, _, _, sa = filter(
                                    lambda x: x[0] == 2,
                                    socket.getaddrinfo(
                                        domain, port, 0, socket.SOCK_STREAM
                                    ),
                                )[0]
                                ip, _ = sa
                                ip_bytes = my_inet_aton(ip)
                                port_bytes = struct.pack("!H", port)
                                self.request.sendall(
                                    b"\x05\x00\x00\x02" + ip_bytes + port_bytes
                                )
                                # Return without actually connecting to domain
                                break
                            # Connect to destination
                            else:
                                dest = self.connect(domain, port, data)

                                # If connected to upstream/destination, let client know
                                dsockname = dest.getsockname()
                                client_ip = dsockname[0]
                                client_port = dsockname[1]
                                ip_bytes = my_inet_aton(client_ip)
                                port_bytes = struct.pack("!H", client_port)
                                self.request.sendall(
                                    b"\x05\x00\x00\x01" + ip_bytes + port_bytes
                                )

                            stage = CONN_ACCEPTED
                        except:
                            logging.exception(
                                "Error when trying to resolve/connect to: %s",
                                (domain, port),
                            )
                            self.request.sendall(b"\x05\x01")
                            raise

            # Starting to forward data
            try:
                if dest:
                    result = self.forward(self.request, dest)
                    if result:
                        logging.debug("Forwarding finished")
                    else:
                        logging.debug("Exception/timeout when forwarding")
            except:
                logging.exception("Error when forwarding")
            finally:
                if dest:
                    dest.close()
                logging.info(
                    "%d bytes out, %d bytes in. Socks5 session finished %s <-> %s.",
                    self.bytes_out,
                    self.bytes_in,
                    self.client_name,
                    self.server_name,
                )
                if self.local_auth and (self.bytes_in or self.bytes_out):
                    self.authority.usage(self.member_id, self.bytes_in + self.bytes_out)
        except Socks5Exception:
            logging.exception("Connection closed")
        except:
            logging.exception("Error when proxying")
            # traceback.print_exc()
        finally:
            try:
                self.request.shutdown(socket.SHUT_RDWR)
            except:
                self.request.close()
            return

    def connect(self, domain, port, data):
        # Connect to upstream instead of destination
        if self.upstream_addr:
            sc = Socks5Client(
                self.upstream_addr,
                self.upstream_username,
                self.upstream_password,
                data,
                enable_ssl=self.upstream_ssl,
            )
            logging.info(
                "Connecting to %s via upstream %s.", domain, self.upstream_addr
            )
            return sc.connect()
        else:
            # Connect to destination directly
            if len(self.allowed_ports) > 0 and port not in self.allowed_ports:
                raise Socks5PortForbidden(port)
            my_ip = self.request.getsockname()[0]
            logging.info("Connecting to %s.", domain)
            return make_connection((domain, port), my_ip)

    def forward(self, client, server):
        """forward data between sockets"""
        self.client_name = client.getpeername()
        self.server_name = server.getpeername()

        while True:
            readables, writeables, exceptions = select.select(
                [client, server], [], [], TIME_OUT
            )

            # exception or timeout
            if exceptions or (readables, writeables, exceptions) == ([], [], []):
                return False

            data = ""

            for readable in readables:
                data = readable.recv(BUF_SIZE)

                if data:
                    if readable == client:
                        self.bytes_out += len(data)
                        server.send(data)
                    else:
                        self.bytes_in += len(data)
                        client.send(data)
                else:
                    return True


class Socks5Client:
    """A socks5 client with optional SSL support"""

    def __init__(
        self,
        addr,
        username="",
        password="",
        data="",
        enable_ssl=True,
        bind_to=None,
        to_upstream=True,
        dns_only=False,
    ):
        """
        :param addr: socket server address tuple
        :param username: username
        :param password: password
        :param data: a tuple of remote address you plan to connect to, or packed data of it.
        :param enable_ssl: if ssl should be enabled
        :param bind_to: ip to bind to for the local socket
        :param to_upstream: if an upstream is used
        :return: established socket or resolved address when dns_only is True
        """
        self.addr = addr
        self.enable_ssl = enable_ssl
        self.username = username.encode("utf-8")
        self.password = password.encode("utf-8")
        self.data = data
        self.bind_to = bind_to
        self.to_upstream = to_upstream
        self.dns_only = dns_only

    def connect(self):
        dest = make_connection(self.addr, self.bind_to, self.to_upstream)
        # SSL enabled
        if dest and self.enable_ssl:
            dest = ssl_context.wrap_socket(dest)

        if not dest:
            trigger_upstream_check()
            raise Socks5ConnectionFailed()

        # Server needs authentication
        if self.username and self.password:
            # Send auth method (username/password auth)
            dest.sendall(b"\x05\x01\x02")
            ans = dest.recv(BUF_SIZE)
            # Method accepted
            if ans == b"\x05\x02":
                name_length = struct.pack("B", len(self.username))
                pass_length = struct.pack("B", len(self.password))
                # Start auth
                dest.sendall(
                    b"\x01" + name_length + self.username + pass_length + self.password
                )
                ans = dest.recv(BUF_SIZE)
                # Auth failed
                if ans != b"\x01\x00":
                    if not ans or ans[1] == AUTH_ERR_SERVER:
                        raise Socks5AuthFailed("An error occurred on server")
                    elif ans[1] == AUTH_ERR_BANDWIDTH:
                        raise Socks5AuthFailed("Bandwidth usage exceeded quota")
                    elif ans[1] == AUTH_ERR_NOPLANFOUND:
                        raise Socks5AuthFailed("Can't find a subscribed plan for user")
                    elif ans[1] == AUTH_ERR_USERNOTFOUND:
                        raise Socks5AuthFailed("User not found or wrong password")
                    else:
                        raise Socks5AuthFailed
            else:
                raise Socks5AuthFailed("No accepted authentication method")
        # No auth needed
        else:
            dest.sendall(b"\x05\x01\x00")
            ans = dest.recv(BUF_SIZE)
            if ans != b"\x05\x00":
                raise Socks5AuthFailed

        if type(self.data) is tuple:
            domain, port = self.data
            domain = domain.encode("utf-8")
            port_str = struct.pack("!H", port)
            len_str = struct.pack("B", len(domain))
            if self.dns_only:
                addr_type = b"\x02"
            else:
                addr_type = b"\x03"
            data = b"\x05\x01\x00" + addr_type + len_str + domain + port_str
        else:
            data = self.data

        dest.sendall(data)
        ans = dest.recv(BUF_SIZE)
        if ans.startswith(b"\x05\x00"):
            if ans[3] == b"\x02":
                return socket.inet_ntoa(ans[4:8])
            else:
                return dest
        else:
            raise Socks5ConnectionFailed



#!/usr/bin/env python

import getopt
import logging.handlers
import sys
from os.path import exists

from .config import FurionConfig as cfg
from .dns import *
from .helpers import *
from .ping import PingHandler
from .servers import *
from .socks5 import *


def usage():
    print("Usage: furion -c <furion.cfg>")


def setup_server():
    config_path = "furion.cfg" if exists("furion.cfg") else "/etc/furion/furion.cfg"

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hc:")
    except getopt.GetoptError:
        usage()
        sys.exit(1)
    for opt, arg in opts:
        if opt == "-c":
            config_path = arg
        else:
            usage()
            sys.exit()

    # Initialize logger
    log_format = "%(asctime)s [%(filename)s:%(lineno)d][%(levelname)s] %(message)s"
    logging.basicConfig(format=log_format, filemode="w")

    logger = logging.getLogger()

    # Initialize config
    if not exists(config_path):
        print("Fatal error: %s is not found, exiting..." % config_path)
        time.sleep(3)
        sys.exit(-1)

    cfg.init(config_path)

    if cfg.log_path:
        formatter = logging.Formatter(log_format)
        log_handler = logging.handlers.RotatingFileHandler(
            cfg.log_path, maxBytes=1024 * 1024 * 10, backupCount=5
        )
        log_handler.setFormatter(formatter)
        logger.addHandler(log_handler)

    logger.setLevel(cfg.log_level)

    try:
        import gevent
        import gevent.monkey

        gevent.monkey.patch_all()
    except ImportError:
        gevent = None

    if gevent:
        logging.info("Using gevent model...")
    else:
        logging.info("Using threading model...")

    if cfg.upstream_list:
        # Setup threads for upstream checking
        thr = threading.Thread(target=run_check, args=(cfg,))
        thr.setDaemon(1)
        thr.start()

        thr = threading.Thread(target=update_upstream, args=(cfg,))
        thr.setDaemon(1)
        thr.start()

    # Start UDP ping server
    if cfg.ping_server:
        ping_svr = PingServer((cfg.local_ip, cfg.ping_server_port), PingHandler)
        thr = threading.Thread(target=ping_svr.serve_forever, args=(5,))
        thr.setDaemon(1)
        thr.start()

    # Start UDP DNS server
    if cfg.dns_server:

        class DNSHandler(DNSQueryHandler, cfg):
            pass

        dns_proxy = DNSServer((cfg.local_ip, cfg.dns_server_port), DNSHandler)
        thr = threading.Thread(target=dns_proxy.serve_forever, args=())
        thr.setDaemon(1)
        thr.start()

    # Re-check upstream every 30 minutes
    thr = threading.Thread(target=check_upstream_repeatedly, args=(1800,))
    thr.setDaemon(1)
    thr.start()

    class FurionHandler(Socks5RequestHandler, cfg):
        pass

    if cfg.local_ssl:
        svr = SecureSocks5Server(
            cfg.pem_path, cfg.local_addr, FurionHandler, cfg.with_systemd
        )
    else:
        svr = Socks5Server(cfg.local_addr, FurionHandler)

    logging.info(
        "Furion server listening on %s, SSL %s, AUTH %s."
        % (
            cfg.local_addr,
            "ON" if cfg.local_ssl else "OFF",
            "ON" if cfg.local_auth else "OFF",
        )
    )

    return svr


def main():
    svr = setup_server()
    try:
        svr.serve_forever()
    except KeyboardInterrupt:
        print("Exiting...")
        svr.server_close()
        sys.exit()


if __name__ == "__main__":
    main()



import socket
import socketserver
import ssl
import threading
from queue import Queue

ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)

socketserver.TCPServer.allow_reuse_address = True

TIME_OUT = 30
POOL_SIZE = 50


class ThreadPoolMixIn(socketserver.ThreadingMixIn):
    """Thread pool mixin"""

    def serve_forever(self, pool_size=POOL_SIZE):
        self.requests = Queue(pool_size)

        for x in range(pool_size):
            t = threading.Thread(target=self.process_request_thread)
            t.setDaemon(1)
            t.start()

        while True:
            self.handle_request()

        self.server_close()

    def process_request_thread(self):
        while True:
            socketserver.ThreadingMixIn.process_request_thread(
                self, *self.requests.get()
            )

    def handle_request(self):
        try:
            request, client_address = self.get_request()
        except socket.error:
            return
        if self.verify_request(request, client_address):
            self.requests.put((request, client_address))


class SecureTCPServer(socketserver.TCPServer):
    """TCP server with SSL"""

    SYSTEMD_FIRST_SOCKET_FD = 3

    def __init__(self, pem_path, server_address, handler_class, with_systemd=False):
        socketserver.BaseServer.__init__(self, server_address, handler_class)

        af, socktype, proto, canonname, sa = socket.getaddrinfo(
            self.server_address[0], self.server_address[1], 0, socket.SOCK_STREAM
        )[0]

        if not with_systemd:
            sock = socket.socket(af, socktype, proto)
        else:
            sock = socket.fromfd(self.SYSTEMD_FIRST_SOCKET_FD, af, socktype)
        # sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(TIME_OUT)

        # Load the certificate
        ssl_context.load_cert_chain(pem_path, pem_path)

        # Don't do handshake on connect for ssl (which will block http://bugs.python.org/issue1251)
        self.socket = ssl_context.wrap_socket(
            sock,
            server_side=True,
            do_handshake_on_connect=False,
        )
        if not with_systemd:
            self.server_bind()
            self.server_activate()


class Socks5Server(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """Threading Socks5 server"""

    pass


class TPSocks5Server(ThreadPoolMixIn, socketserver.TCPServer):
    """Thread Pool Socks5 server"""

    pass


class SecureSocks5Server(socketserver.ThreadingMixIn, SecureTCPServer):
    """Secure Socks5 server"""

    pass


class TPSecureSocks5Server(ThreadPoolMixIn, SecureTCPServer):
    """Thread Pool Secure Socks5 server"""

    pass


class PingServer(ThreadPoolMixIn, socketserver.UDPServer):
    """UDP Ping server"""

    pass


class DNSServer(socketserver.ThreadingMixIn, socketserver.UDPServer):
    """UDP DNS Proxy"""

    pass


# Test server
# svr = PingServer(('0.0.0.0', 8888), PingHandler)
# svr.serve_forever(5)



import logging
import math
import select
import socket
import socketserver
import time


def ping(addr, count=20, timeout=1):
    """UDP ping client"""
    # print "--- PING %s:%d ---" % addr
    results = []
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    for i in range(count):
        ts = time.time()
        data = "PING %d %f %s" % (i, ts, "#" * 480)
        data = data.encode("utf-8")
        sock.sendto(data, addr)
        readables, writeables, exceptions = select.select([sock], [], [], timeout)
        # exception
        if exceptions:
            time.sleep(1)
            continue
        # timeout
        if (readables, writeables, exceptions) == ([], [], []):
            continue
        if readables:
            ret = readables[0].recv(512)
            if ret == data:
                time_spent = (time.time() - ts) * 1000
                results.append(time_spent)
                # print '%d bytes from %s:%d, seq=%d time=%.3f ms' % (len(data), addr[0], addr[1], i, time_spent)

    received = len(results)
    missing = count - received
    loss = count - received
    # print "--- %s:%d ping statistics---" % addr
    # print "%d packets transmitted, %d packets received, %.1f%% packet loss" % (count, received, float(loss)*100/count)
    logging.debug(
        "ping %s result: %d transmitted, %d received, %.1f%% loss",
        addr,
        count,
        received,
        float(loss) * 100 // count,
    )
    if received != 0:
        min_val = min(results)
        max_val = max(results)
        avg = sum(results) // count
        stddev = math.sqrt(sum([(x - avg) ** 2 for x in results]) // received)
        # print "round-trip min/avg/max/stddev = %.3f/%.3f/%.3f/%.3f" % (min_val, avg, max_val, stddev)
        logging.debug(
            "ping %s min/avg/max/stddev = %.3f/%.3f/%.3f/%.3f",
            addr,
            min_val,
            avg,
            max_val,
            stddev,
        )
        return missing * 500 + avg
    else:
        return float("inf")


class PingHandler(socketserver.BaseRequestHandler):
    """UDP Ping server handler"""

    def handle(self):
        data = self.request[0].strip()
        sock = self.request[1]
        sock.sendto(data, self.client_address)
        # print data


# Test client
# import threading
# for x in range(10):
# threading.Thread(target = ping, args = (('192.3.21.149', 8888),)).start()



import os, os.path
import sys
import wx
import wx.adv
import threading
import subprocess

from furion.furion import setup_server

ID_ICON_TIMER = wx.NewIdRef()
OPEN_CONFIG=wx.NewIdRef()
SHOW_LOGS=wx.NewIdRef()


def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    base_path = getattr(sys, '_MEIPASS', os.path.dirname(os.path.abspath(__file__)))
    return os.path.join(base_path, relative_path)


class FurionTaskBarIcon(wx.adv.TaskBarIcon):

    def __init__(self, parent):  
        wx.adv.TaskBarIcon.__init__(self)  
        self.parentApp = parent  
        self.icon = wx.Icon(resource_path("furion.ico"), wx.BITMAP_TYPE_ICO)  
        self.CreateMenu()  
        self.SetIconImage()

    def CreateMenu(self):  
        self.Bind(wx.adv.EVT_TASKBAR_RIGHT_UP, self.ShowMenu)  
        self.Bind(wx.EVT_MENU, self.parentApp.openConfig, id=OPEN_CONFIG)  
        self.Bind(wx.EVT_MENU, self.parentApp.showLogs, id=SHOW_LOGS)  
        self.menu=wx.Menu()  
        self.menu.Append(OPEN_CONFIG, "Open Config File")
        # self.menu.Append(SHOW_LOGS, "Show Logs")            
        self.menu.AppendSeparator()  
        self.menu.Append(wx.ID_EXIT, "Exit Furion")

    def ShowMenu(self,event):  
        self.PopupMenu(self.menu)

    def SetIconImage(self, mail=False):  
        self.SetIcon(self.icon)


class FurionFrame(wx.Frame):

    def __init__(self, parent, id, title):  
        wx.Frame.__init__(self, parent, -1, title, size = (1, 1),  
            style=wx.FRAME_NO_TASKBAR|wx.NO_FULL_REPAINT_ON_RESIZE)

        self.tbicon = FurionTaskBarIcon(self)  
        self.tbicon.Bind(wx.EVT_MENU, self.exitApp, id=wx.ID_EXIT)   
        self.Show(True)
        self.svr = None
        self.thr = threading.Thread(target=self.runServer)
        self.thr.setDaemon(1)
        self.thr.start()

    def runServer(self):
        self.svr = setup_server()
        self.svr.serve_forever()


    def exitServer(self):
        self.svr.shutdown()
        self.Close()

    def exitApp(self, event):
        self.tbicon.RemoveIcon()  
        self.tbicon.Destroy()  
        self.exitServer()
        

    def openConfig(self, event):
        fpath = 'furion.cfg'
        if sys.platform.startswith('darwin'):
            subprocess.call(('open', fpath))
        elif os.name == 'nt': # For Windows
            os.startfile(fpath)
        elif os.name == 'posix': # For Linux, Mac, etc.
            subprocess.call(('xdg-open', fpath))

    def showLogs(self, event):  
        pass


def main(argv=None):
    app = wx.App(False)
    frame = FurionFrame(None, -1, ' ')
    frame.Show(False)  
    app.MainLoop()

if __name__ == '__main__':
    main()



import json
import sys
import time
from configparser import ConfigParser
from io import StringIO
from os.path import abspath, basename, dirname, exists, join

from .helpers import get_upstream_from_central
from .simpleauth import SimpleAuth

default_config = """
[main]
local_ip = 127.0.0.1
local_port = 11080
with_systemd = off
local_ssl = off
pem_path =
local_auth = off
allowed_ports = 22,53,80,443
ping_server = off
ping_server_port = 17777
dns_server = on
dns_server_port = 15353
remote_tcp_dns = 8.8.4.4
log_level = 20
log_path =

[upstream]
central_url =
autoupdate_upstream_list = off
update_frequency = weekly
upstream_list_path = upstream.json
# upstream_ip =
# upstream_port = 443
# upstream_ssl = on
# upstream_auth = on
# upstream_username =
# upstream_password =

[plugin]
auth_plugin =
"""


class FurionConfig(object):
    last_update = 0

    @classmethod
    def get_path(cls, path):
        if not path:
            return None
        if not exists(path) and basename(path) == path:
            return join(cls.config_dir, path)
        else:
            return path

    @classmethod
    def init(cls, path):
        default_cfg = StringIO(default_config)

        cls.config = ConfigParser()
        cls.config.read_file(default_cfg)
        cls.config.read_file(open(path))
        cls.config_dir = dirname(abspath(path))

        auth_plugin = cls.config.get("plugin", "auth_plugin")

        cls.authority = None
        if auth_plugin == "simpleauth":
            cls.password_path = cls.get_path(
                cls.config.get("simpleauth", "password_path")
            )
            cls.authority = SimpleAuth(cls.password_path)

        cls.local_ip = cls.config.get("main", "local_ip")
        cls.local_port = cls.config.getint("main", "local_port")
        cls.with_systemd = cls.config.getboolean("main", "with_systemd")

        cls.local_ssl = cls.config.getboolean("main", "local_ssl")
        cls.local_auth = cls.config.getboolean("main", "local_auth")
        cls.pem_path = cls.get_path(cls.config.get("main", "pem_path"))
        if cls.pem_path and not exists(cls.pem_path):
            print('Fatal error: pem "%s" cannot be found.' % cls.pem_path)
            time.sleep(3)
            sys.exit(-1)
        ports = cls.config.get("main", "allowed_ports").strip()
        if ports == "all" or ports == "":
            cls.allowed_ports = []
        else:
            cls.allowed_ports = [int(port) for port in ports.split(",")]

        cls.ping_server = cls.config.getboolean("main", "ping_server")
        cls.ping_server_port = cls.config.getint("main", "ping_server_port")
        cls.dns_server = cls.config.getboolean("main", "dns_server")
        cls.dns_server_port = cls.config.getint("main", "dns_server_port")
        cls.remote_tcp_dns = cls.config.get("main", "remote_tcp_dns")
        cls.log_level = cls.config.getint("main", "log_level")
        cls.log_path = cls.get_path(cls.config.get("main", "log_path"))

        cls.central_url = cls.config.get("upstream", "central_url")
        cls.autoupdate_upstream_list = cls.config.getboolean(
            "upstream", "autoupdate_upstream_list"
        )
        cls.update_frequency = cls.config.get("upstream", "update_frequency")
        cls.upstream_list_path = cls.get_path(
            cls.config.get("upstream", "upstream_list_path")
        )

        cls.upstream_list = None
        if exists(cls.upstream_list_path):
            cls.upstream_list = json.loads(open(cls.upstream_list_path).read())[
                "upstream_list"
            ]
        elif cls.autoupdate_upstream_list:
            get_upstream_from_central(cls)
        # cls.upstream_ip = cls.config.get('upstream', 'upstream_ip')
        # cls.upstream_port = cls.config.getint('upstream', 'upstream_port')
        # cls.upstream_ssl = cls.config.getboolean('upstream', 'upstream_ssl')
        # cls.upstream_auth = cls.config.getboolean('upstream', 'upstream_auth')
        # cls.upstream_username = cls.config.get('upstream', 'upstream_username')
        # cls.upstream_password = cls.config.get('upstream', 'upstream_password')

        cls.local_addr = (cls.local_ip, cls.local_port)
        cls.upstream_addr = None
        cls.upstream_ping = None
        # cls.upstream_addr = (cls.upstream_ip, cls.upstream_port) if cls.upstream_ip else None



import json
import logging
import os
import socket
import ssl
import threading
import time
from queue import Queue
from urllib.request import Request, urlopen

from .ping import ping

MIN_INTERVAL = 30
UPSTREAM_TIMEOUT = 10
CONN_TIMEOUT = 5

# NoticeQueue is used for triggering an upstream check
NoticeQueue = Queue(1)
# Alive upstream servers are put into this queue
UpstreamQueue = Queue(100)

ssl_context = ssl.create_default_context()
ssl_context.check_hostname = False
ssl_context.verify_mode = ssl.CERT_NONE


def make_connection(addr, bind_to=None, to_upstream=False):
    """
    Make TCP connection and return socket.

    :param addr: (domain, port) tuple to connect to
    :param bind_to: ip address to bind to
    :param to_upstream: if the destination is an upstream server
    :return: socket
    """

    domain, port = addr
    if to_upstream:
        timeout = UPSTREAM_TIMEOUT
    else:
        timeout = CONN_TIMEOUT

    for res in socket.getaddrinfo(domain, port, 0, socket.SOCK_STREAM):
        af, socktype, proto, canonname, sa = res
        client = None
        try:
            client = socket.socket(af, socktype, proto)
            client.settimeout(timeout)
            # Specify outgoing IP on a multihome server
            # if bind_to and bind_to != '127.0.0.1':
            # client.bind((bind_to, 0))
            client.connect(sa)
            return client
        except Exception as e:
            if client is not None:
                client.close()
            logging.debug(
                "Error occurred when making connection to dest %s: %s", addr, e
            )
            return None


def set_upstream(cfg, upstream):
    cfg.upstream_addr = upstream["ip"], upstream["port"]
    cfg.upstream_auth = upstream["auth"]
    cfg.upstream_ssl = upstream["ssl"]
    cfg.upstream_username = upstream["username"]
    cfg.upstream_password = upstream["password"]


def get_upstream_from_central(cfg, timing="now"):
    if timing == "weekly":
        st = os.stat(cfg.upstream_list_path)
        if time.time() - st.st_mtime < 86400 * 7:
            logging.info("Ignore request to get upstream from central...")
            return None
    if cfg.central_url and cfg.autoupdate_upstream_list:
        try:
            logging.info("Fetching upstream from central...")
            jsonstr = urlopen(cfg.central_url).read()
            jsonstr = jsonstr.decode("utf-8")
            cfg.upstream_list = json.loads(jsonstr)["upstream_list"]
        except Exception as e:
            logging.exception("Failed to fetch upstream from central")
        else:
            logging.info("Saving upstream list...")
            open(cfg.upstream_list_path, "w").write(jsonstr)
            return True
    else:
        logging.fatal("No central_url is configured or autoupdate is off.")
        return False


def run_check(cfg):
    """Check alive upstream servers"""
    if not cfg.upstream_list:
        get_upstream_from_central(cfg)
    else:
        # set a default upstream if none is set already
        if not cfg.upstream_addr:
            set_upstream(cfg, cfg.upstream_list[0])

        if cfg.update_frequency == "start":
            get_upstream_from_central(cfg)
        elif cfg.update_frequency == "weekly":
            get_upstream_from_central(cfg, "weekly")

    while True:
        ts = NoticeQueue.get()
        diff = ts - cfg.last_update
        if cfg.last_update == 0 or diff > MIN_INTERVAL:
            logging.info(
                "Last check %d seconds ago, checking for live upstream...", diff
            )
            for upstream in cfg.upstream_list:
                t = threading.Thread(target=check_alive, args=(upstream,))
                t.start()


def check_alive(upstream):
    dest = None
    try:
        addr = (upstream["ip"], upstream["port"])
        dest = make_connection(addr, None, True)
        # SSL enabled
        if dest and upstream["ssl"]:
            dest = ssl_context.wrap_socket(dest)
            logging.debug("Upstream %s is ALIVE", addr)

        if not dest:
            logging.debug("Upstream %s is DEAD: Connection failed", addr)
    except Exception as e:
        if dest is not None:
            dest.close()
        logging.debug("Upstream %s is DEAD: %s", addr, e)
        return
    try:
        ping_port = upstream["ping_port"] if "ping_port" in upstream else 17777
        score = ping((upstream["ip"], ping_port))
        upstream["ping"] = score
        UpstreamQueue.put((time.time(), upstream))
    except Exception as e:
        logging.debug("Ping to %s failed: %s", (upstream["ip"], ping_port), e)


def update_upstream(cfg):
    while True:
        ts, upstream = UpstreamQueue.get()
        if cfg.last_update == 0 or ts - cfg.last_update > MIN_INTERVAL:
            logging.info("Setting upstream to: %s", (upstream["ip"], upstream["port"]))
            cfg.last_update = ts
            set_upstream(cfg, upstream)
        else:
            logging.debug("Upstream %s is not used", (upstream["ip"], upstream["port"]))


def trigger_upstream_check():
    logging.debug("Triggering upstream check...")
    NoticeQueue.put_nowait(time.time())


def check_upstream_repeatedly(seconds):
    while True:
        trigger_upstream_check()
        time.sleep(seconds)


# Some versions of windows don't have socket.inet_pton
if hasattr(socket, "inet_pton"):

    def my_inet_aton(ip_string):
        af, _, _, _, _ = socket.getaddrinfo(ip_string, 80, 0, socket.SOCK_STREAM)[0]
        return socket.inet_pton(af, ip_string)
else:
    my_inet_aton = socket.inet_aton


def hexstring(s):
    return " ".join(["%02X" % ord(c) for c in s])


# http://stackoverflow.com/a/14620633/1349791
class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self



from .socks5 import AUTH_ERR_USERNOTFOUND, AUTH_SUCCESSFUL


class SimpleAuth:
    def __init__(self, path):
        self.users = {}
        for line in open(path):
            line = line.strip()
            if line:
                user, password = line.split(" ")
                self.users[user] = password

    def auth(self, username, password):
        if self.users.get(username, "") == password:
            return 0, AUTH_SUCCESSFUL
        else:
            return 0, AUTH_ERR_USERNOTFOUND

    def usage(self, member_id, bytes):
        pass



#!/usr/bin/env python3
import argparse
import subprocess
import sys


def run_command(command):
    process = subprocess.run(command, shell=True, check=True)
    return process.returncode


def main():
    parser = argparse.ArgumentParser(description="Publish package to PyPI or TestPyPI")
    parser.add_argument("version", help="Version to publish (e.g., 0.2.0)")
    parser.add_argument(
        "--production",
        "-p",
        action="store_true",
        help="Publish to PyPI instead of TestPyPI",
    )
    args = parser.parse_args()

    version = args.version
    if not version.startswith("v"):
        version = f"v{version}"

    # Default repository is TestPyPI
    twine_command = "twine upload --repository testpypi dist/*"
    if args.production:
        twine_command = "twine upload dist/*"

    commands = [
        f"git tag {version}",
        f"git push origin {version}",
        "rm -rf dist/ build/ *.egg-info",
        "python -m build",
        twine_command,
    ]

    for command in commands:
        print(f"Executing: {command}")
        try:
            run_command(command)
        except subprocess.CalledProcessError as e:
            print(f"Error executing command: {command}")
            print(f"Error: {e}")
            sys.exit(1)

    repo = "PyPI" if args.production else "TestPyPI"
    print(f"Successfully published version {version} to {repo}")


if __name__ == "__main__":
    main()
